/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwitrans.h

    Purpose:

        This file implements the translation routines that abstracts the
        differences between various ASIC families.
        

--*/

#ifndef __hwitrans_h__
#define __hwitrans_h__

typedef ACIT_VOID (* PHWI_PROCESS_CQ) (PACIT_CONTROLLER_WS pcontroller,
                                       ACIT_UEXACT32 message_id);

typedef ACIT_VOID (* PHWI_TRANSLATE_ISOL_CQE) (struct _HWI_CONTROLLER_WS * phwi,
                                               PACIT_VOID hw_cqe,
                                               struct _ACIT_SOL_CQE * drv_cqe);

typedef ACIT_VOID (* PHWI_TRANSLATE_DPDU_CQE) (struct _HWI_CONTROLLER_WS * phwi,
                                               PACIT_VOID hw_cqe,
                                               struct _ACIT_DPDU_CQE * drv_cqe);

typedef ACIT_VOID (*PHWI_FILL_IO_REQUEST) (struct _HWI_CONTROLLER_WS * phwi,
                                           PACIT_IO_HANDLE);

typedef ACIT_VOID (*PHWI_FILL_TMF_REQUEST) (struct _HWI_CONTROLLER_WS * phwi,
                                           PACIT_TMF_HANDLE);

typedef ACIT_VOID (*PHWI_FILL_LOGOUT_REQUEST) (struct _HWI_CONTROLLER_WS * phwi,
                                               PACIT_LOGOUT_HANDLE);

typedef ACIT_VOID (*PHWI_FILL_NOP_REQUEST) (struct _HWI_CONTROLLER_WS * phwi,
                                           PACIT_NOP_HANDLE,
                                           PACIT_NOP_HANDLE);

typedef struct _HWI_TRANSLATION_TABLE {

    PHWI_PROCESS_CQ             hwi_process_cq;
    PHWI_TRANSLATE_ISOL_CQE     hwi_trans_isol_cqe;
    PHWI_TRANSLATE_DPDU_CQE     hwi_trans_dpdu_cqe;
    PHWI_FILL_IO_REQUEST        hwi_fill_io_request;
    PHWI_FILL_TMF_REQUEST       hwi_fill_tmf_request;
    PHWI_FILL_LOGOUT_REQUEST    hwi_fill_logout_request;
    PHWI_FILL_NOP_REQUEST       hwi_fill_nop_request;

} HWI_TRANSLATION_TABLE, *PHWI_TRANSLATION_TABLE;

// Function pointer access macros
#define HWI_PROCESS_CQ(pc, msg) (pc->phwi_ws->translation_routines.hwi_process_cq)(pc, msg)
#define HWI_TRANSLATE_ISOL_CQE(ph, s, d) (ph->translation_routines.hwi_trans_isol_cqe)(ph,s,d)
#define HWI_TRANSLATE_DPDU_CQE(ph, s, d) (ph->translation_routines.hwi_trans_dpdu_cqe)(ph,s,d)
#define HWI_FILL_IO_REQUEST(ph, pioh) (ph->translation_routines.hwi_fill_io_request)(ph, pioh)
#define HWI_FILL_TMF_REQUEST(ph, ptmf) (ph->translation_routines.hwi_fill_tmf_request)(ph, ptmf)
#define HWI_FILL_LOGOUT_REQUEST(ph, plog) (ph->translation_routines.hwi_fill_logout_request)(ph, plog)
#define HWI_FILL_NOP_REQUEST(ph, pnop, popt) (ph->translation_routines.hwi_fill_nop_request)(ph, pnop, popt)

ACIT_VOID
hwi_init_translation_routines(
    struct _HWI_CONTROLLER_WS * phwi);

ACIT_VOID
hwi_translate_be2_isol_cqe(
    struct _HWI_CONTROLLER_WS * phwi,
    PACIT_VOID              phw_cqe,
    PACIT_SOL_CQE           pdrv_cqe);

ACIT_VOID
hwi_translate_be4_isol_cqe(
    struct _HWI_CONTROLLER_WS * phwi,
    PACIT_VOID              phw_cqe,
    PACIT_SOL_CQE           pdrv_cqe);

ACIT_VOID
hwi_translate_be2_dpdu_cqe(
    struct _HWI_CONTROLLER_WS * phwi,
    PACIT_VOID              phw_cqe,
    PACIT_DPDU_CQE          pdrv_cqe);

ACIT_VOID
hwi_translate_be4_dpdu_cqe(
    struct _HWI_CONTROLLER_WS * phwi,
    PACIT_VOID              phw_cqe,
    PACIT_DPDU_CQE          pdrv_cqe);

ACIT_VOID
hwi_process_be2_cq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_VOID
hwi_process_be4_cq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_UEXACT32
hwi_fill_be2_io_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
hwi_fill_be2_logout_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_LOGOUT_HANDLE     plogout_handle);

ACIT_UEXACT32
hwi_fill_be2_tmf_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_UEXACT32
hwi_fill_be2_nop_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_NOP_HANDLE        pnop_handle,
    PACIT_NOP_HANDLE        poptional_handle);

ACIT_UEXACT32
hwi_fill_be4_io_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
hwi_fill_be4_logout_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_LOGOUT_HANDLE     plogout_handle);

ACIT_UEXACT32
hwi_fill_be4_tmf_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_UEXACT32
hwi_fill_be4_nop_request(
    struct _HWI_CONTROLLER_WS * phwi_controller, 
    PACIT_NOP_HANDLE        pnop_handle,
    PACIT_NOP_HANDLE        poptional_handle);

ACIT_VOID
hwi_mark_be2_wrb_for_invalidate(
    struct _WRB_HANDLE *    pwrb_handle);

ACIT_VOID
hwi_mark_be4_wrb_for_invalidate(
    struct _WRB_HANDLE *    pwrb_handle);

#endif
