/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!
@file
   hwiwrb.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/initiator/acit/hwiwrb.h,v 1.8 2006/03/23 20:40:44 prabalm Exp $
*/
#ifndef __hwiwrb_h__
#define __hwiwrb_h__

/*
 * The following routines are actually part of HWI API, but are
 * defined by HWIWRB for efficiency
 */

ACIT_UEXACT32
hwi_submit_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PWRB_HANDLE             pwrb_handle,
    ACIT_UEXACT32           cri);

ACIT_UEXACT32
hwi_submit_drvr_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              pwrb_handle,
    ACIT_UEXACT32           cri);

ACIT_VOID
hwi_complete_cmd(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    PACIT_SOL_CQE           psol);

ACIT_UEXACT32
hwi_complete_drvr_msgs(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cid,
    ACIT_UEXACT32           wrb_index);

ACIT_VOID
hwi_complete_cmd_error(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           wrb_index,
    ACIT_UEXACT16           code);

/* HWI API end */

PWRB_HANDLE
hwi_alloc_wrb_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri);

ACIT_VOID
hwi_free_wrb_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri,
    PWRB_HANDLE             pwrb_handle);

PSGL_HANDLE
hwi_alloc_io_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller);

ACIT_VOID
hwi_free_io_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    PSGL_HANDLE             psgl_handle);

PSGL_HANDLE
hwi_alloc_eh_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller);

ACIT_VOID
hwi_free_eh_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    PSGL_HANDLE             psgl_handle);

ACIT_VOID
hwi_get_wrb_mem(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cid,
    PACIT_UEXACT32          pwrb_mem_index,
    PACIT_UEXACT32          poffset,
    PACIT_UEXACT32          psize);


ACIT_UEXACT32
hwi_write_be2_wrb_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           sgl_index);

ACIT_UEXACT32
hwi_write_be4_wrb_sgl(
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           sgl_index);

ACIT_UEXACT32
hwi_write_be2_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list);

ACIT_UEXACT32
hwi_write_be4_sgl(
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list);

#ifdef ACIT_CFG_CHAINED_SGL

ACIT_VOID
_hwi_fill_chained_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_DC_LIST           psgh_list,
    PSGL_HANDLE             psgl_handle,
    PACIT_DC_LIST           list_position,
    PACIT_SG_LIST           pdata_sg_list);

ACIT_UEXACT32
hwi_write_be2_chained_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_DC_LIST           psgh_list,
    PACIT_SG_LIST           pdata_sg_list);

ACIT_UEXACT32
hwi_write_be4_chained_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_DC_LIST           psgh_list,
    PACIT_SG_LIST           pdata_sg_list);

#endif

ACIT_UEXACT32
hwi_write_be2_large_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list);

ACIT_UEXACT32
hwi_write_be4_large_sgl(
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list);

ACIT_VOID
_hwi_init_wrb_handle(
    PACIT_CONTROLLER_WS     pcontroller);


ACIT_VOID
hwi_init_wrb_handle(
    PACIT_CONTROLLER_WS      pcontroller);

ACIT_VOID
hwi_init_sgl_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_reserve_nop_sgl(
    PACIT_CONTROLLER_WS     pcontroller);

#if defined (ACIT_CFG_DMA_TEST) || defined (ACIT_CFG_CHAINED_SGL)
#define FILL_SGE(sge, _len, _addr_lo, _addr_hi, _offset, _last) {   \
                        ACIT_PUT_LE32(sge->len, _len);              \
                        ACIT_PUT_LE32(sge->addr_lo, _addr_lo);      \
                        ACIT_PUT_LE32(sge->addr_hi, _addr_hi);      \
                        ACIT_PUT_LE32(sge->sge_offset, _offset);    \
                        ACIT_PUT_LE32(sge->sge_offset, _offset);    \
                        sge->rsvd0 = 0;                             \
                        sge->rsvd1 = 0;                             \
                        sge->last_sge = _last;                      \
                        }
#else
#define FILL_SGE(sge, _len, _addr_lo, _addr_hi, _offset, _last) {   \
                        ACIT_PUT_LE32(sge->len, _len);              \
                        ACIT_PUT_LE32(sge->addr_lo, _addr_lo);      \
                        ACIT_PUT_LE32(sge->addr_hi, _addr_hi);      \
                        ACIT_PUT_LE32(sge->sge_offset, _offset);    \
                        sge->last_sge = _last;                      \
                        }
#endif


#ifdef ACIT_CFG_CHAINED_SGL

// Only 29 bit CRA is enough. Basically WCON ignores the lower 3 bits
// (offset within a SGE) and left shifts the CRA by 3 bits.
#define FILL_CHAIN_SGE(sge, _cra) {                                 \
    sge->rsvd0[0] = 0;                                              \
    ACIT_PUT_LE32(sge->cra, ((_cra.dw >> 3) & 0x1FFFFFFF));         \
    sge->rsvd1 = 0;                                                 \
    sge->rsvd2 = 0;                                                 \
    sge->rsvd3[0] = 0;                                              \
    sge->type = WCON_NEXT_SGE_T_V2;                                 \
    sge->last = 0;                                                  \
    }

#endif

INLINE
ACIT_VOID
_hwi_write_sgl(
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list);

ACIT_VOID
_hwi_write_large_sgl(
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list);

#endif
