/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwitrans.c

    Purpose:

        This file implements the translation routines that abstracts the
        differences between various ASIC families.
        

--*/

#include "acitpch.h"

ACIT_VOID
hwi_init_translation_routines(
    struct _HWI_CONTROLLER_WS * phwi)
{
    PACIT_CONTROLLER_WS pcontroller;

    pcontroller = phwi->pacit_ws;

    // Switch the tables based on ASIC generation. All generations after BE2 and BE3
    // will use the new definitions
    if(IS_ASIC_BE2_OR_BE3(pcontroller))
    {
        phwi->translation_routines.hwi_process_cq = (PHWI_PROCESS_CQ)hwi_process_be2_cq;
        phwi->translation_routines.hwi_trans_isol_cqe = (PHWI_TRANSLATE_ISOL_CQE)hwi_translate_be2_isol_cqe;
        phwi->translation_routines.hwi_trans_dpdu_cqe = (PHWI_TRANSLATE_DPDU_CQE)hwi_translate_be2_dpdu_cqe;
        phwi->translation_routines.hwi_fill_io_request = (PHWI_FILL_IO_REQUEST)hwi_fill_be2_io_request;
        phwi->translation_routines.hwi_fill_tmf_request = (PHWI_FILL_TMF_REQUEST)hwi_fill_be2_tmf_request;
        phwi->translation_routines.hwi_fill_logout_request = (PHWI_FILL_LOGOUT_REQUEST)hwi_fill_be2_logout_request;
        phwi->translation_routines.hwi_fill_nop_request = (PHWI_FILL_NOP_REQUEST)hwi_fill_be2_nop_request;
    }
    else
    {
        phwi->translation_routines.hwi_process_cq = (PHWI_PROCESS_CQ)hwi_process_be4_cq;
        phwi->translation_routines.hwi_trans_isol_cqe = (PHWI_TRANSLATE_ISOL_CQE)hwi_translate_be4_isol_cqe;
        phwi->translation_routines.hwi_trans_dpdu_cqe = (PHWI_TRANSLATE_DPDU_CQE)hwi_translate_be4_dpdu_cqe;
        phwi->translation_routines.hwi_fill_io_request = (PHWI_FILL_IO_REQUEST)hwi_fill_be4_io_request;
        phwi->translation_routines.hwi_fill_tmf_request = (PHWI_FILL_TMF_REQUEST)hwi_fill_be4_tmf_request;
        phwi->translation_routines.hwi_fill_logout_request = (PHWI_FILL_LOGOUT_REQUEST)hwi_fill_be4_logout_request;
        phwi->translation_routines.hwi_fill_nop_request = (PHWI_FILL_NOP_REQUEST)hwi_fill_be4_nop_request;
    }
}

ACIT_VOID
hwi_translate_be2_isol_cqe(
    PHWI_CONTROLLER_WS      phwi,
    PACIT_VOID              phw_cqe,
    PACIT_SOL_CQE           pdrv_cqe)
{
    PI_SOL_CQE              pbe2_cqe;

    pbe2_cqe = (PI_SOL_CQE)phw_cqe;

    pdrv_cqe->hw_sts = (ACIT_UEXACT8)pbe2_cqe->hw_sts;
    pdrv_cqe->i_flags.byte = (ACIT_UEXACT8)pbe2_cqe->i_flags;
    pdrv_cqe->i_resp = (ACIT_UEXACT8)pbe2_cqe->i_resp;
    pdrv_cqe->i_sts = (ACIT_UEXACT8)pbe2_cqe->i_sts;
    pdrv_cqe->cid = (ACIT_UEXACT16)pbe2_cqe->cid;
    pdrv_cqe->wrb_index = (ACIT_UEXACT16)pbe2_cqe->wrb_index;
    pdrv_cqe->i_exp_cmd_sn = pbe2_cqe->i_exp_cmd_sn;
    pdrv_cqe->i_res_cnt = pbe2_cqe->i_res_cnt;
    pdrv_cqe->i_cmd_wnd = (ACIT_UEXACT8)pbe2_cqe->i_cmd_wnd;
    pdrv_cqe->s = (ACIT_UEXACT8)pbe2_cqe->s;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("%p - HWSts : 0x%x iSts : 0x%x iRsp : 0x%x iFlags : 0x%x SBit : %d iResCnt : %d\n",
        phw_cqe, pbe2_cqe->hw_sts, pbe2_cqe->i_sts, pbe2_cqe->i_resp, pbe2_cqe->i_flags, pbe2_cqe->s, pbe2_cqe->i_res_cnt));

    return;
}

ACIT_VOID
hwi_translate_be4_isol_cqe(
    PHWI_CONTROLLER_WS      phwi,
    PACIT_VOID              phw_cqe,
    PACIT_SOL_CQE           pdrv_cqe)
{
    PI_SKH_SOL_CQE_V2       pbe4_cqe;

    pbe4_cqe = (PI_SKH_SOL_CQE_V2)phw_cqe;

    pdrv_cqe->hw_sts = (ACIT_UEXACT8)pbe4_cqe->hw_sts;
    pdrv_cqe->i_flags.bits.o = (ACIT_UEXACT8)pbe4_cqe->o;
    pdrv_cqe->i_flags.bits.u = (ACIT_UEXACT8)pbe4_cqe->u;

    // The BE4 CQE contains either the status or the response.
    // As per RFC 3720, the Status field in the PDU is used to report the SCSI
    // status of the command as per SAM2 and is valid only if the Response code
    // is Command Completed at the target (0x00)
    // If cmd_complete is set, then i_sts_rsp is SCSI Status. Else it is iSCSI Error
    // Response.
    if(pbe4_cqe->cmd_complete == 1)
    {
        pdrv_cqe->i_sts = (ACIT_UEXACT8)pbe4_cqe->i_sts_rsp;
        // Status is now Command completed at the target
        pdrv_cqe->i_resp = SCSI_SVC_RESP_GOOD;
    }
    else
    {
        pdrv_cqe->i_resp = (ACIT_UEXACT8)pbe4_cqe->i_sts_rsp;
        // -1 one is to tell there was no status
        pdrv_cqe->i_sts = (ACIT_UEXACT8)-1;
        OSM_ASSERT_MSG((pdrv_cqe->i_resp),("Response cannot be zero when cmd_complete is zero\n"));
    }

    pdrv_cqe->cid = (ACIT_UEXACT16)pbe4_cqe->cid;
    pdrv_cqe->wrb_index = (ACIT_UEXACT16)pbe4_cqe->wrb_index;
    pdrv_cqe->i_exp_cmd_sn = pbe4_cqe->i_exp_cmd_sn;
    pdrv_cqe->i_res_cnt = pbe4_cqe->i_res_cnt;
    pdrv_cqe->i_cmd_wnd = (ACIT_UEXACT8)pbe4_cqe->i_cmd_wnd;
    pdrv_cqe->s = (ACIT_UEXACT8)pbe4_cqe->s;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("%p - HWSts : 0x%x O : %d U : %d CmdCmp : %d iStsRsp : %d SBit : %d iResCnt : 0x%x\n",
        phw_cqe, pbe4_cqe->hw_sts, pbe4_cqe->o, pbe4_cqe->u, pbe4_cqe->cmd_complete, pbe4_cqe->i_sts_rsp, pbe4_cqe->s, pbe4_cqe->i_res_cnt));

    return;
}

ACIT_VOID
hwi_translate_be2_dpdu_cqe(
    PHWI_CONTROLLER_WS      phwi,
    PACIT_VOID              phw_cqe,
    PACIT_DPDU_CQE          pdrv_cqe)
{
    PI_T_DPDU_CQE           pbe2_cqe;

    pbe2_cqe = (PI_T_DPDU_CQE)phw_cqe;

    pdrv_cqe->db_addr_hi = pbe2_cqe->db_addr_hi;
    pdrv_cqe->db_addr_lo = pbe2_cqe->db_addr_lo;
    pdrv_cqe->DataPlacementLength = (ACIT_UEXACT16)pbe2_cqe->DataPlacementLength;
    pdrv_cqe->index = (ACIT_UEXACT16)pbe2_cqe->index;
    pdrv_cqe->cid = (ACIT_UEXACT16)pbe2_cqe->cid;
    pdrv_cqe->num_cons = (ACIT_UEXACT8)pbe2_cqe->num_cons;
    pdrv_cqe->code = (ACIT_UEXACT8)pbe2_cqe->code;

    return;
}

ACIT_VOID
hwi_translate_be4_dpdu_cqe(
    PHWI_CONTROLLER_WS      phwi,
    PACIT_VOID              phw_cqe,
    PACIT_DPDU_CQE          pdrv_cqe)
{
    PI_T_SKH_DPDU_CQE_V2    pbe4_cqe;

    pbe4_cqe = (PI_T_SKH_DPDU_CQE_V2)phw_cqe;

    pdrv_cqe->db_addr_hi = pbe4_cqe->db_addr_hi;
    pdrv_cqe->db_addr_lo = pbe4_cqe->db_addr_lo;
    pdrv_cqe->DataPlacementLength = (ACIT_UEXACT16)pbe4_cqe->DataPlacementLength;
    pdrv_cqe->index = (ACIT_UEXACT16)pbe4_cqe->index;
    pdrv_cqe->cid = (ACIT_UEXACT16)pbe4_cqe->cid;
    pdrv_cqe->num_cons = (ACIT_UEXACT8)pbe4_cqe->num_cons;
    pdrv_cqe->code = (ACIT_UEXACT8)pbe4_cqe->code;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("%p - Lo : 0x%x Hi : 0x%x Len : %d Idx : %d Cons : %d Cid : 0x%x Code : 0x%x\n",
        phw_cqe, pbe4_cqe->db_addr_lo, pbe4_cqe->db_addr_hi, pbe4_cqe->DataPlacementLength,
        pbe4_cqe->index, pbe4_cqe->num_cons, pbe4_cqe->cid, pbe4_cqe->code));

    return;
}

ACIT_VOID
hwi_process_be2_cq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{            
    PHWI_CONTROLLER_WS      phwi;
    SA_RING                 *ring;
    PISCSI_CQE              cqe;
    PI_SOL_CQE              sol;
    PI_T_DMSG_CQE           dmsg;
    ACIT_UEXACT32           num_processed = 0; // num processed this time
    ACIT_UEXACT32           count = 0;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_cq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);

    ring = &pcontroller->partition[message_id].cq;

    cqe = (PISCSI_CQE) sa_ring_current (ring); // CQ entry to process
    ACIT_UPDATE_CQ_INT_COUNT(pcontroller, message_id);

    while (cqe->valid && (count < pcontroller->params.max_cqe_per_isr)) 
    {
        ACIT_UPDATE_CQ_PROCESSED_COUNT(pcontroller, message_id);

        if (num_processed >= 32) 
        {
            // Occassionally free space in CQ for the hardware to write
            // more entries without enabling the interrupt yet.
            hwi_ring_cq_db (pcontroller, ring->id, num_processed, 0, 0);
            num_processed = 0;
        }     

        switch(cqe->code)
        {
        case SOL_CMD_COMPLETE:
            {    
                ACIT_SOL_CQE sol_cqe;
                HWI_TRANSLATE_ISOL_CQE(phwi, (PACIT_VOID)cqe, &sol_cqe);
                hwi_complete_cmd(phwi, message_id, &sol_cqe);
                ACIT_UPDATE_ISCSI_CMD_INT_COUNT(pcontroller, message_id);
            }

            break;

        case DRIVERMSG_NOTIFY:

            dmsg = (PI_T_DMSG_CQE)cqe;
            // Free the WRB index in the chain until this WRB.  
            hwi_complete_drvr_msgs(phwi, dmsg->cid, dmsg->wrb_index);
            ACIT_UPDATE_DRIVER_MSG_INT_COUNT(pcontroller, message_id);

            break;

        case UNSOL_HDR_NOTIFY:
        case UNSOL_DATA_NOTIFY:
            {    
                ACIT_DPDU_CQE dpdu_cqe;
                HWI_TRANSLATE_DPDU_CQE(phwi, (PACIT_VOID)cqe, &dpdu_cqe);
                hwi_process_default_pdu_ring(pcontroller, &dpdu_cqe);
                ACIT_UPDATE_DFT_PDU_INT_COUNT(pcontroller, message_id);
            }

            break;

        case CXN_INVALIDATE_INDEX_NOTIFY:

            // FixmeBE2: Not implemented!!
            OSM_ASSERT(0);

            break;

        case CMD_INVALIDATED_NOTIFY:
        case CXN_INVALIDATE_NOTIFY:

            // These can be ignored
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Ignoring CQ Error notification for cmd/cxn invalidate\n"));
            break;

        case SOL_CMD_KILLED_DATA_DIGEST_ERR:
        case CMD_KILLED_INVALID_STATSN_RCVD:
        case CMD_KILLED_INVALID_R2T_RCVD:
        case CMD_CXN_KILLED_LUN_INVALID:
        case CMD_CXN_KILLED_ICD_INVALID:
        case CMD_CXN_KILLED_ITT_INVALID:
        case CMD_CXN_KILLED_SEQ_OUTOFORDER:
        case CMD_CXN_KILLED_INVALID_DATASN_RCVD:

            // Though some of the above code say CMD_CXN, it only affects the command in 
            // initiator mode. Target mode affects the connection, hence the _CXN..

            sol = (PI_SOL_CQE)cqe;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error notification for cmd.. code 0x%x cid 0x%x\n",
                                                cqe->code, sol->cid));

            iscsi_update_statistics(pcontroller, sol->cid, cqe->code, sol->wrb_index, 
                                     32, (PACIT_VOID)sol);

            hwi_complete_cmd_error(phwi, 
                                    message_id,
                                    (ACIT_UEXACT16)sol->cid, 
                                    (ACIT_UEXACT16)sol->wrb_index, 
                                    (ACIT_UEXACT16)cqe->code);

            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        case UNSOL_DATA_DIGEST_ERROR_NOTIFY:

            sol = (PI_SOL_CQE)cqe;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Digest error on def pdu ring, dropping..\n"));

            // Increment default PDU producer index and post it back
            iscsi_update_statistics(pcontroller, cqe->cid, cqe->code, sol->wrb_index,
                                     32, (PACIT_VOID)sol);
            //
            // Bug 24334: This CQE does not consume a DPDU buffer and hence we need
            // not post the DPDU back. Mark the connection so the next DPDU on the
            // data ring need to be discarded because its got a digest error
            //
            iscsi_process_unsol_data_digest_err_notify(pcontroller, cqe->cid);
            
            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        case CXN_KILLED_PDU_SIZE_EXCEEDS_DSL:
        case CXN_KILLED_BURST_LEN_MISMATCH:
        case CXN_KILLED_AHS_RCVD:
        case CXN_KILLED_HDR_DIGEST_ERR:
        case CXN_KILLED_UNKNOWN_HDR:
        case CXN_KILLED_STALE_ITT_TTT_RCVD:
        case CXN_KILLED_INVALID_ITT_TTT_RCVD:
        case CXN_KILLED_FIN_RCVD:
        case CXN_KILLED_BAD_UNSOL_PDU_RCVD:
        case CXN_KILLED_BAD_WRB_INDEX_ERROR:
        case CXN_KILLED_OVER_RUN_RESIDUAL:
        case CXN_KILLED_UNDER_RUN_RESIDUAL:
        case CXN_KILLED_CMND_DATA_NOT_ON_SAME_CONN:

            sol = (PI_SOL_CQE)cqe;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error 0x%x, resetting CID 0x%x...\n", cqe->code, cqe->cid));
            iscsi_update_statistics(pcontroller, cqe->cid, cqe->code, sol->wrb_index,
                                     32, (PACIT_VOID)sol);

            iscsi_rst_send_notify(pcontroller, cqe->cid);
            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        case CXN_KILLED_RST_SENT:        
        case CXN_KILLED_RST_RCVD:
        case CXN_KILLED_TCP_PROT_ERR:

            sol = (PI_SOL_CQE)cqe;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error 0x%x, reset received/sent on CID 0x%x...\n", cqe->code, cqe->cid));
            iscsi_update_statistics(pcontroller, cqe->cid, cqe->code, sol->wrb_index,
                                     32, (PACIT_VOID)sol);

            iscsi_rst_recv_notify(pcontroller, cqe->cid);
            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        default:
            OSM_ASSERT(0);
            pcontroller->partition[message_id].interrupt_stats.unknown_cqes++;
            break;

        }

        // Mark this CQ_ENTRY as invalid
        cqe->valid = FALSE;

        // Circularly increment the consumer index, and move to next CQ_ENTRY
        // to process.
        cqe = (PISCSI_CQE) sa_ring_next (ring);
        
        num_processed++;
        count++;
    }

    // Update on-chip ring index.
    if (num_processed > 0) 
    {
        hwi_ring_cq_db (pcontroller, ring->id, num_processed, 1, 0);
    }

    if(count > 64)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt64++;
    else if (count > 32)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt32++;
    else if (count > 16)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt16++;
    else
        pcontroller->partition[message_id].interrupt_stats.cqe_lt16++;

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_cq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);
    return;
}

ACIT_VOID
hwi_process_be4_cq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{            
    PHWI_CONTROLLER_WS      phwi;
    SA_RING                 *ring;
    PISCSI_SKH_CQE_V2       cqe;
    PI_SKH_SOL_CQE_V2       sol;
    PI_T_SKH_DMSG_CQE_V2    dmsg;
    ACIT_UEXACT32           num_processed = 0; // num processed this time
    ACIT_UEXACT32           count = 0;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_cq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);

    ring = &pcontroller->partition[message_id].cq;

    cqe = (PISCSI_SKH_CQE_V2) sa_ring_current (ring); // CQ entry to process
    ACIT_UPDATE_CQ_INT_COUNT(pcontroller, message_id);

    while (cqe->valid && (count < pcontroller->params.max_cqe_per_isr)) 
    {
        ACIT_UPDATE_CQ_PROCESSED_COUNT(pcontroller, message_id);

        if (num_processed >= 32) 
        {
            // Occassionally free space in CQ for the hardware to write
            // more entries without enabling the interrupt yet.
            hwi_ring_cq_db (pcontroller, ring->id, num_processed, 0, 0);
            num_processed = 0;
        }     

        switch(cqe->code)
        {
        case SOL_CMD_COMPLETE:
            {    
                ACIT_SOL_CQE sol_cqe;
                HWI_TRANSLATE_ISOL_CQE(phwi, (PACIT_VOID)cqe, &sol_cqe);
                hwi_complete_cmd(phwi, message_id, &sol_cqe);
                ACIT_UPDATE_ISCSI_CMD_INT_COUNT(pcontroller, message_id);
            }

            break;

        case DRIVERMSG_NOTIFY:

            dmsg = (PI_T_SKH_DMSG_CQE_V2)cqe;
            // Free the WRB index in the chain until this WRB.  
            hwi_complete_drvr_msgs(phwi, dmsg->cid, dmsg->wrb_index);
            ACIT_UPDATE_DRIVER_MSG_INT_COUNT(pcontroller, message_id);

            break;

        case UNSOL_HDR_NOTIFY:
        case UNSOL_DATA_NOTIFY:
            {    
                ACIT_DPDU_CQE dpdu_cqe;
                HWI_TRANSLATE_DPDU_CQE(phwi, (PACIT_VOID)cqe, &dpdu_cqe);
                hwi_process_default_pdu_ring(pcontroller, &dpdu_cqe);
                ACIT_UPDATE_DFT_PDU_INT_COUNT(pcontroller, message_id);
            }

            break;

        case CXN_INVALIDATE_INDEX_NOTIFY:

            // FixmeBE2: Not implemented!!
            OSM_ASSERT(0);

            break;

        case CMD_INVALIDATED_NOTIFY:
        case CXN_INVALIDATE_NOTIFY:

            // These can be ignored
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Ignoring CQ Error notification for cmd/cxn invalidate\n"));
            break;

        case SOL_CMD_KILLED_DATA_DIGEST_ERR:
        case CMD_KILLED_INVALID_STATSN_RCVD:
        case CMD_KILLED_INVALID_R2T_RCVD:
        case CMD_CXN_KILLED_LUN_INVALID:
        case CMD_CXN_KILLED_ICD_INVALID:
        case CMD_CXN_KILLED_ITT_INVALID:
        case CMD_CXN_KILLED_SEQ_OUTOFORDER:
        case CMD_CXN_KILLED_INVALID_DATASN_RCVD:

            // Though some of the above code say CMD_CXN, it only affects the command in 
            // initiator mode. Target mode affects the connection, hence the _CXN..

            sol = (PI_SKH_SOL_CQE_V2)cqe;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error notification for cmd.. code 0x%x cid 0x%x\n",
                                                sol->code, sol->cid));

            iscsi_update_statistics(pcontroller, sol->cid, sol->code, sol->wrb_index,
                                     32, (PACIT_VOID)sol);

            hwi_complete_cmd_error(phwi, 
                                    message_id,
                                    (ACIT_UEXACT16)sol->cid, 
                                    (ACIT_UEXACT16)sol->wrb_index, 
                                    (ACIT_UEXACT16)cqe->code);

            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        case UNSOL_DATA_DIGEST_ERROR_NOTIFY:

            sol = (PI_SKH_SOL_CQE_V2)cqe;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Digest error on def pdu ring, dropping..\n"));

            // Increment default PDU producer index and post it back
            iscsi_update_statistics(pcontroller, sol->cid, sol->code, sol->wrb_index,
                                     32, (PACIT_VOID)sol);

            //
            // Bug 24334: This CQE does not consume a DPDU buffer and hence we need
            // not post the DPDU back. Mark the connection so the next DPDU on the
            // data ring need to be discarded because its got a digest error
            //
            iscsi_process_unsol_data_digest_err_notify(pcontroller, sol->cid);
            
            pcontroller->eh_stats.cq_errors[sol->code]++;

            break;

        case CXN_KILLED_PDU_SIZE_EXCEEDS_DSL:
        case CXN_KILLED_BURST_LEN_MISMATCH:
        case CXN_KILLED_AHS_RCVD:
        case CXN_KILLED_HDR_DIGEST_ERR:
        case CXN_KILLED_UNKNOWN_HDR:
        case CXN_KILLED_STALE_ITT_TTT_RCVD:
        case CXN_KILLED_INVALID_ITT_TTT_RCVD:
        case CXN_KILLED_FIN_RCVD:
        case CXN_KILLED_BAD_UNSOL_PDU_RCVD:
        case CXN_KILLED_BAD_WRB_INDEX_ERROR:
        case CXN_KILLED_OVER_RUN_RESIDUAL:
        case CXN_KILLED_UNDER_RUN_RESIDUAL:
        case CXN_KILLED_CMND_DATA_NOT_ON_SAME_CONN:

            sol = (PI_SKH_SOL_CQE_V2)cqe;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error 0x%x, resetting CID 0x%x...\n", sol->code, sol->cid));
            iscsi_update_statistics(pcontroller, sol->cid, sol->code, sol->wrb_index,
                                     32, (PACIT_VOID)sol);

            iscsi_rst_send_notify(pcontroller, sol->cid);
            pcontroller->eh_stats.cq_errors[sol->code]++;

            break;

        case CXN_KILLED_RST_SENT:        
        case CXN_KILLED_RST_RCVD:
        case CXN_KILLED_TCP_PROT_ERR:

            sol = (PI_SKH_SOL_CQE_V2)cqe;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error 0x%x, reset received/sent on CID 0x%x...\n", sol->code, sol->cid));
            iscsi_update_statistics(pcontroller, sol->cid, sol->code, sol->wrb_index,
                                     32, (PACIT_VOID)sol);

            iscsi_rst_recv_notify(pcontroller, sol->cid);
            pcontroller->eh_stats.cq_errors[sol->code]++;

            break;

        default:
            OSM_ASSERT(0);
            pcontroller->partition[message_id].interrupt_stats.unknown_cqes++;
            break;

        }

        // Mark this CQ_ENTRY as invalid
        cqe->valid = FALSE;

        // Circularly increment the consumer index, and move to next CQ_ENTRY
        // to process.
        cqe = (PISCSI_SKH_CQE_V2) sa_ring_next (ring);
        
        num_processed++;
        count++;
    }

    // Update on-chip ring index.
    if (num_processed > 0) 
    {
        hwi_ring_cq_db (pcontroller, ring->id, num_processed, 1, 0);
    }

    if(count > 64)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt64++;
    else if (count > 32)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt32++;
    else if (count > 16)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt16++;
    else
        pcontroller->partition[message_id].interrupt_stats.cqe_lt16++;

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_cq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);
    return;
}

ACIT_UEXACT32
hwi_fill_be2_io_request(
    PHWI_CONTROLLER_WS          phwi_controller, 
    PACIT_IO_HANDLE             pio_handle)
{
    PWRB_HANDLE                 pwrb_handle;
    PISCSI_WRB                  pwrb;
    PACIT_IO_REQUEST            pio_request = pio_handle->pio->pio_request;

    pwrb_handle = (PWRB_HANDLE)pio_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pbe2_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET(pwrb, 0, sizeof(ISCSI_WRB));

    // TODO - handle multi-level LUNs here

    OSM_ASSERT(!(pio_request->lun.u.lun_leveled[1] |
                 pio_request->lun.u.lun_leveled[2] |
                 pio_request->lun.u.lun_leveled[3]));

    // 14 bits of LUN info, from the first level in bigendian format
    ACIT_PUT_BE16(pwrb->lun, pio_request->lun.u.lun_leveled[0]);

    if(pio_request->flags.outbound_data) 
    {
        pwrb->type    = INI_WR_CMD;
		pwrb->dsp     = 1;

        // for writes, use r2t_buf_offset field to communicate number of sg
        // elements in order to turn on the EP firmware's fast path
        //pwrb->r2t_buf_offset = pio_request->psg_list->number_elements;
    }
    else
    {
        pwrb->type = INI_RD_CMD;
    }
#if ACIT_CFG_INTERRUPT_REDIRECTION
    // Merely by compile flag alone, we assume we can stuff in the cq_id into
    // the ISCSI_WRB. The BE2 firmware has no such field however. But that
    // field on BE2 is a reserved member, so we go ahead and fill it anyway. 
    // Saves a couple of checks
    pwrb->cq_id             = pio_handle->cq_id;
#endif
    pwrb->wrb_idx           = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb       = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->r2t_exp_dtl, pio_handle->exp_xfer_length);
    ACIT_PUT_LE32(pwrb->cmdsn_itt, pio_handle->cmd_sn);

#ifdef ACIT_CFG_CHAINED_SGL
        hwi_write_be2_chained_sgl(phwi_controller->pacit_ws,
                            pio_handle, // for debug
                            pwrb,
                            &pio_handle->pio->bhs_pa,
                            pio_handle->pio->bhs_len,
                            &pio_handle->psglh_list,
                            pio_request->psg_list);

#else
    // Fill the SGs.
    // split logic is not required upto 64KB.
    if(pio_handle->exp_xfer_length <= BE_MAX_BYTES_PER_FRAG)
    {
        hwi_write_be2_sgl(pwrb,
                      &pio_handle->pio->bhs_pa,
                      pio_handle->pio->bhs_len,
                      pio_handle->psgl_handle, 
                      pio_request->psg_list);
    }
    else
    {
        hwi_write_be2_large_sgl(pwrb,
                            &pio_handle->pio->bhs_pa,
                            pio_handle->pio->bhs_len,
                            pio_handle->psgl_handle,
                            pio_request->psg_list);
    }
#endif

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("IO WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], pio_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_be2_logout_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_WRB              pwrb;

    pwrb_handle = plogout_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pbe2_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // Logout is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit. 
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx    = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->cmdsn_itt, plogout_handle->cmd_sn);
    // put the logout completion on the first CQ
#if ACIT_CFG_INTERRUPT_REDIRECTION
    pwrb->cq_id = HWI_GET_CQ_ID(phwi_controller->pacit_ws, 0);
#endif

    /*
     * Fill the SGs.
     */
    hwi_write_be2_sgl(pwrb,
                  &plogout_handle->bhs_pa,
                  plogout_handle->bhs_len,
                  plogout_handle->psgl_handle, 
                  ACIT_NULL);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("Logout WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], plogout_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_be2_tmf_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_WRB              pwrb;

    pwrb_handle = ptmf_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pbe2_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // TMF is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx    = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->cmdsn_itt, ptmf_handle->cmd_sn);

    if(ptmf_handle->tmf_function == ACIT_TMF_ABORT_TASK)
    {
        pwrb->ref_sgl_icd_idx = (ptmf_handle->pio_handle->itt & 0xfff);
    }
    // put the tmf completion on the first CQ
#if ACIT_CFG_INTERRUPT_REDIRECTION
    pwrb->cq_id = HWI_GET_CQ_ID(phwi_controller->pacit_ws, 0);
#endif

    /*
     * Fill the SGs.
     */
    hwi_write_be2_sgl(pwrb,
                  &ptmf_handle->bhs_pa,
                  ptmf_handle->bhs_len,
                  ptmf_handle->psgl_handle, 
                  ACIT_NULL);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("TMF WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], ptmf_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_be2_nop_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_NOP_HANDLE        pnop_handle,
    PACIT_NOP_HANDLE        poptional_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_WRB              pwrb;

    pwrb_handle = pnop_handle->pwrb_handle;
    pwrb = (PISCSI_WRB)pwrb_handle->pbe2_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // NOPOut is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit. 
    // Note!! If there is a payload that needs to go out, then 
    // dsp bit should be set
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx     = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    pwrb->dmsg        = ACIT_TRUE;  // ack based completion
    ACIT_PUT_LE32(pwrb->cmdsn_itt, pnop_handle->cmd_sn);

    if(poptional_handle)
    {
        /*
         * Fill the SGs.
         */
        hwi_write_be2_sgl(pwrb,
                      &pnop_handle->bhs_pa,
                      pnop_handle->bhs_len,
                      pnop_handle->psgl_handle, 
                      ACIT_NULL);
    }
    else
    {
        hwi_write_be2_wrb_sgl(pwrb, &pnop_handle->bhs_pa, phwi_controller->nop_icd);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("NOP WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], pnop_handle->cri));


    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_be4_io_request(
    PHWI_CONTROLLER_WS          phwi_controller, 
    PACIT_IO_HANDLE             pio_handle)
{
    PWRB_HANDLE                 pwrb_handle;
    PISCSI_SKH_WRB_V2           pwrb;
    PACIT_IO_REQUEST            pio_request = pio_handle->pio->pio_request;

    pwrb_handle = (PWRB_HANDLE)pio_handle->pwrb_handle;
    pwrb = (PISCSI_SKH_WRB_V2)pwrb_handle->pbe4_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET(pwrb, 0, sizeof(ISCSI_SKH_WRB_V2));

    // TODO - handle multi-level LUNs here

    OSM_ASSERT(!(pio_request->lun.u.lun_leveled[1] |
                 pio_request->lun.u.lun_leveled[2] |
                 pio_request->lun.u.lun_leveled[3]));

    // 14 bits of LUN info, from the first level in bigendian format
    ACIT_PUT_BE16(pwrb->lun, pio_request->lun.u.lun_leveled[0]);

    if(pio_request->flags.outbound_data) 
    {
        pwrb->type    = INI_WR_CMD;
		pwrb->dsp     = 1;

        // for writes, use r2t_buf_offset field to communicate number of sg
        // elements in order to turn on the EP firmware's fast path
        //pwrb->r2t_buf_offset = pio_request->psg_list->number_elements;
    }
    else
    {
        pwrb->type = INI_RD_CMD;
    }
#if ACIT_CFG_INTERRUPT_REDIRECTION
    // Merely by compile flag alone, we assume we can stuff in the cq_id into
    // the ISCSI_WRB. The BE2 firmware has no such field however. But that
    // field on BE2 is a reserved member, so we go ahead and fill it anyway. 
    // Saves a couple of checks
    pwrb->cq_id             = pio_handle->cq_id;
#endif
    pwrb->wrb_idx           = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb       = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->r2t_exp_dtl, pio_handle->exp_xfer_length);
    ACIT_PUT_LE32(pwrb->cmdsn_itt, pio_handle->cmd_sn);

#ifdef ACIT_CFG_CHAINED_SGL
    if(phwi_controller->pacit_ws->params.dma_test && pio_request->flags.inbound_data)
    {
        hwi_write_be4_chained_sgl(phwi_controller->pacit_ws,
                            pio_handle, // for debug
                            pwrb,
                            &pio_handle->pio->bhs_pa,
                            pio_handle->pio->bhs_len,
                            &pio_handle->psglh_list,
                            pio_request->psg_list);
    }
    else
    {
        PSGL_HANDLE psgl_handle;
        PACIT_DC_LIST list_position;
        list_position = pio_handle->psglh_list.next;
        psgl_handle = STRUCT_BASE(SGL_HANDLE, link, list_position);

        // Fill the SGs.
        // split logic is not required upto 64KB.
        if(pio_handle->exp_xfer_length <= BE_MAX_BYTES_PER_FRAG)
        {
            hwi_write_be4_sgl(pwrb,
                          &pio_handle->pio->bhs_pa,
                          pio_handle->pio->bhs_len,
                          psgl_handle, 
                          pio_request->psg_list);
        }
        else
        {
            hwi_write_be4_large_sgl(pwrb,
                                &pio_handle->pio->bhs_pa,
                                pio_handle->pio->bhs_len,
                                psgl_handle,
                                pio_request->psg_list);
        }
    }
#else
    // Fill the SGs.
    // split logic is not required upto 64KB.
    if(pio_handle->exp_xfer_length <= BE_MAX_BYTES_PER_FRAG)
    {
        hwi_write_be4_sgl(pwrb,
                      &pio_handle->pio->bhs_pa,
                      pio_handle->pio->bhs_len,
                      pio_handle->psgl_handle, 
                      pio_request->psg_list);
    }
    else
    {
        hwi_write_be4_large_sgl(pwrb,
                            &pio_handle->pio->bhs_pa,
                            pio_handle->pio->bhs_len,
                            pio_handle->psgl_handle,
                            pio_request->psg_list);
    }
#endif

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("IO WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], pio_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_be4_logout_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_SKH_WRB_V2       pwrb;

    pwrb_handle = plogout_handle->pwrb_handle;
    pwrb = (PISCSI_SKH_WRB_V2)pwrb_handle->pbe4_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // Logout is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit. 
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx    = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->cmdsn_itt, plogout_handle->cmd_sn);
    // put the logout completion on the first CQ
#if ACIT_CFG_INTERRUPT_REDIRECTION
    pwrb->cq_id = HWI_GET_CQ_ID(phwi_controller->pacit_ws, 0);
#endif

    /*
     * Fill the SGs.
     */
    hwi_write_be4_sgl(pwrb,
                  &plogout_handle->bhs_pa,
                  plogout_handle->bhs_len,
                  plogout_handle->psgl_handle, 
                  ACIT_NULL);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("Logout WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], plogout_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_be4_tmf_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_SKH_WRB_V2       pwrb;

    pwrb_handle = ptmf_handle->pwrb_handle;
    pwrb = (PISCSI_SKH_WRB_V2)pwrb_handle->pbe4_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // TMF is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx    = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    ACIT_PUT_LE32(pwrb->cmdsn_itt, ptmf_handle->cmd_sn);

    if(ptmf_handle->tmf_function == ACIT_TMF_ABORT_TASK)
    {
        pwrb->ref_sgl_icd_idx = (ptmf_handle->pio_handle->itt & 0xfff);
    }
    // put the tmf completion on the first CQ
#if ACIT_CFG_INTERRUPT_REDIRECTION
    pwrb->cq_id = HWI_GET_CQ_ID(phwi_controller->pacit_ws, 0);
#endif

    /*
     * Fill the SGs.
     */
    hwi_write_be4_sgl(pwrb,
                  &ptmf_handle->bhs_pa,
                  ptmf_handle->bhs_len,
                  ptmf_handle->psgl_handle, 
                  ACIT_NULL);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("TMF WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], ptmf_handle->cri));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_fill_be4_nop_request(
    PHWI_CONTROLLER_WS      phwi_controller, 
    PACIT_NOP_HANDLE        pnop_handle,
    PACIT_NOP_HANDLE        poptional_handle)
{
    PWRB_HANDLE             pwrb_handle;
    PISCSI_SKH_WRB_V2       pwrb;

    pwrb_handle = pnop_handle->pwrb_handle;
    pwrb = (PISCSI_SKH_WRB_V2)pwrb_handle->pbe4_wrb;

    OSM_ASSERT (pwrb);
    OSM_MEMSET (pwrb, 0, sizeof (*pwrb));

    // NOPOut is sent with INI_WR_CMD type and no dsp bit is
    // set since there's no data to transmit. 
    // Note!! If there is a payload that needs to go out, then 
    // dsp bit should be set
    pwrb->type = INI_WR_CMD;
    pwrb->wrb_idx     = pwrb_handle->wrb_index;
    pwrb->ptr2NextWrb = pwrb->wrb_idx;
    pwrb->dmsg        = ACIT_TRUE;  // ack based completion
    ACIT_PUT_LE32(pwrb->cmdsn_itt, pnop_handle->cmd_sn);

    if(poptional_handle)
    {
        /*
         * Fill the SGs.
         */
        hwi_write_be4_sgl(pwrb,
                      &pnop_handle->bhs_pa,
                      pnop_handle->bhs_len,
                      pnop_handle->psgl_handle, 
                      ACIT_NULL);
    }
    else
    {
        hwi_write_be4_wrb_sgl(pwrb, &pnop_handle->bhs_pa, phwi_controller->nop_icd);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("NOP WRB: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x cri 0x%x\n", 
            pwrb->dw[0], pwrb->dw[1], pwrb->dw[2], pwrb->dw[3], pwrb->dw[4], pwrb->dw[5], pwrb->dw[6], pwrb->dw[7], pnop_handle->cri));


    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_mark_be4_wrb_for_invalidate(
    struct _WRB_HANDLE *    pwrb_handle)
{
    pwrb_handle->pbe4_wrb->invld = 1;
}

ACIT_VOID
hwi_mark_be2_wrb_for_invalidate(
    struct _WRB_HANDLE *    pwrb_handle)
{
    pwrb_handle->pbe2_wrb->invld = 1;
}
