/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

///*++
//
//    Filename:
//
//        hwiwrb.c
//
//    Purpose:
//
//        This file implements functionality to interact with the WRB data
//        structures. These structures are used to interface with the BE ASIC.
//        
//
//--*/
//
//
#include "acitpch.h"

ACIT_VOID
hwi_unitt_init_sgl_handle(
    PACIT_CONTROLLER_WS pcontroller);

extern
ACIT_VOID
hwi_complete_cmd(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    PACIT_SOL_CQE           psol);

//
// We could have implemented a function pointer for the doorbell request since
// BE4 onwards, the DB format extends the wrb_cid member to accommodate more
// bits. The _V2 version supports upto 16K-1 CID value. But since only the
// reserved bits are modified in V2 and BE2/BE3 will never use a CID > 1024, we
// use the same _V2 version and don't use a different function pointer.
//
ACIT_UEXACT32
hwi_submit_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PWRB_HANDLE             pwrb_handle,
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    WRB_POST_DB_V2          wrb_post_db; // Use _V2 version always.
    ACIT_UEXACT32           wrb_index;
    PHWI_WRB_CONTEXT        pwrb_context;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    wrb_index = ((PWRB_HANDLE)(pwrb_handle))->wrb_index;

    //OSM_ASSERT(!(((PWRB_HANDLE)(pwrb_handle))->count));
    //((PWRB_HANDLE)(pwrb_handle))->count++;

    //
    // WRBs are outstanding so we need to update the tail pointer.  There
    // is a chance that the tail WRB is:
    //   1. Actually the tail and requires linking.
    //   2. Already freed and sitting in the free list.
    //   3. The same as the current WRB that we are posting.
    // In case 2 & 3 the linking is not required but it will not hurt
    // anything.  It is done in all cases to avoid synchronizing the
    // tailWrb variable between the completion and posting routines. In
    // a full duplex storport driver, they are not synchronized by default.
    //

    pwrb_context = &phwi_controller->pwrb_context[cri];

    if(pwrb_context->plast_wrb)
    {
        pwrb_context->plast_wrb->ptr2NextWrb = wrb_index;
        pwrb_context->plast_wrb = ((PWRB_HANDLE)(pwrb_handle))->pbe2_wrb;
    }
    else
    {
        pwrb_context->plast_wrb = ((PWRB_HANDLE)(pwrb_handle))->pbe2_wrb;
    }

    //
    // Format doorbell. This will post the next command.
    //
	memset(&wrb_post_db,0,sizeof(wrb_post_db));
    wrb_post_db.wrb_cid = pwrb_context->cid;
    wrb_post_db.wrb_index = wrb_index;
    wrb_post_db.numberPosted = 1;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("WRB Post Doorbell:0x%x", wrb_post_db.dw));

    PdWrite(pcontroller, pwrb_context->doorbell_offset, wrb_post_db.dw);
    ACIT_UPDATE_TXULP_DB_COUNT(pcontroller, pwrb_context->ulp);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_submit_drvr_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              pwrb_handle,
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_h = (PWRB_HANDLE)pwrb_handle;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    pwrb_context = &phwi_controller->pwrb_context[cri];

    LIST_ADDTO_TAIL(&pwrb_h->link,
                    &pwrb_context->wrb_handle_drvr_list);

    ACIT_UPDATE_DRVR_MSG_SENT_COUNT(pcontroller);

    hwi_submit_request(pcontroller,
                       pwrb_handle,
                       cri);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_complete_cmd(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    PACIT_SOL_CQE           psol)
{
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_handle;
    PWRB_HANDLE             pwrb_handle_base;

    // Find the command object.
    ASSERT (psol->wrb_index < phwi_controller->pacit_ws->params.num_cmd_per_cxn);

    pwrb_context = &phwi_controller->pwrb_context[hwi_cri_from_cid(phwi_controller->pacit_ws, psol->cid)];

    pwrb_handle_base = pwrb_context->pwrb_handle_base;
    pwrb_handle = (PWRB_HANDLE)pwrb_handle_base + psol->wrb_index;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("0x%p : HWSts %d ISts %x IRsp %x IFlgs %x ResCnt %d S %d\n",
        psol, psol->hw_sts, psol->i_sts, psol->i_resp, psol->i_flags, psol->i_res_cnt, psol->s));

    switch(pwrb_handle->type)
    {
    case HWH_TYPE_IO: {
        ACIT_BIT_SET(pwrb_handle->pio_handle->flags, FLAG_IOH_IO_COMPLETION);

#if ACIT_CFG_INTERRUPT_REDIRECTION
        OSM_ASSERT(pwrb_handle->pio_handle->message_id == message_id);
#endif

        phwi_controller->pacit_ws->partition[message_id].dic.ios_per_sec++;

        ap_complete_io(phwi_controller->pacit_ws, 
                       pwrb_handle->pio_handle, 
                       psol);

        }
        break;

    case HWH_TYPE_LOGOUT:

        ap_complete_logout(phwi_controller->pacit_ws, 
                           pwrb_handle->plogout_handle,
                           (ACIT_UEXACT8)psol->i_resp,
                           0);  // no timeout

        break;

    case HWH_TYPE_TMF:
        
        ap_complete_tmf(phwi_controller->pacit_ws, 
                        pwrb_handle->ptmf_handle,
                        (ACIT_UEXACT8)psol->i_resp,
                        (ACIT_UEXACT8)psol->i_cmd_wnd,
                        psol->i_exp_cmd_sn);

        break;

    case HWH_TYPE_NOP:
        /* This is only in case initiator sent out a NOPOut and gets
         * back a NOPIn response. Async NOPIns end up in default PDU
         * queue
         */
        ap_complete_nop_in(phwi_controller->pacit_ws, pwrb_handle->pnop_handle);
        break;

    default:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Invalid command completion WRB Index 0x%x CID 0x%x\n",
                                            psol->wrb_index, psol->cid));

        OSM_ASSERT(0);
        phwi_controller->pacit_ws->partition[message_id].interrupt_stats.unknown_cmds++;
        break;
    }
}

ACIT_UEXACT32
hwi_complete_drvr_msgs(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cid,
    ACIT_UEXACT32           wrb_index)
{
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_handle;
    PWRB_HANDLE             pwrb_handle_base = ACIT_NULL;
    PACIT_DC_LIST           plink;
    ACIT_UEXACT32           this_wrb_index, count = 0;

    /*
     * Walk the WRB handle list until we've found the WRB
     * whose index matches with wrb_index passed
     */
    pwrb_context = &phwi_controller->pwrb_context[hwi_cri_from_cid(phwi_controller->pacit_ws, cid)];
    pwrb_handle_base = pwrb_context->pwrb_handle_base;

    do
    {
        OSM_ASSERT(IS_LIST_NOT_EMPTY(&pwrb_context->wrb_handle_drvr_list));

        LIST_REMOVEFROM_HEAD (&plink, 
                              &pwrb_context->wrb_handle_drvr_list);

        pwrb_handle = STRUCT_BASE(WRB_HANDLE,
                                  link, 
                                  plink);

        OSM_ASSERT(pwrb_handle);
        OSM_ASSERT(pwrb_handle->type == HWH_TYPE_NOP);

        this_wrb_index = pwrb_handle->wrb_index;

        ap_free_nop_resources(phwi_controller->pacit_ws, pwrb_handle->pnop_handle);
        count++;
    }
    while(this_wrb_index != wrb_index);

    ACIT_UPDATE_DRVR_MSG_RECVD_COUNT(phwi_controller->pacit_ws, count);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_complete_cmd_error(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           message_id,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           wrb_index,
    ACIT_UEXACT16           code)
{
    PACIT_CONTROLLER_WS     pcontroller;
    PHWI_WRB_CONTEXT        pwrb_context;
    PWRB_HANDLE             pwrb_handle;
    PWRB_HANDLE             pwrb_handle_base;

    pcontroller = phwi_controller->pacit_ws;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("CQ Error 0x%x, CMD kill received for WRB 0x%x on CID 0x%x...\n"), 
                                        code, wrb_index, cid));

    // Find the command object.
    ASSERT (wrb_index < pcontroller->params.num_cmd_per_cxn);

    pwrb_context = &phwi_controller->pwrb_context[hwi_cri_from_cid(phwi_controller->pacit_ws, cid)];

    pwrb_handle_base = pwrb_context->pwrb_handle_base;
    pwrb_handle = (PWRB_HANDLE)pwrb_handle_base + wrb_index;

    if(pwrb_handle->type != HWH_TYPE_INVALID)
    {
        OSM_ASSERT(pwrb_handle->pio_handle);

        if(ACIT_IS_BIT_CLR(pwrb_handle->pio_handle->flags, FLAG_IOH_INVALIDATE_PENDING) &&
            ACIT_IS_BIT_CLR(pwrb_handle->pio_handle->flags, FLAG_IOH_INVALIDATE_COMPLETE))
        {
            switch(pwrb_handle->type)
            {
            case HWH_TYPE_IO:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on IOH 0x%p in target 0x%p error, will trigger abort after invalidation...\n"),
                    pwrb_handle->pio_handle, pwrb_handle->pio_handle->ptarget));

                ap_complete_io_error(pcontroller, 
                                     pwrb_handle->pio_handle,
                                     FLAG_IOH_CQ_ERROR);

                break;

            case HWH_TYPE_LOGOUT:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on LOH 0x%p in target 0x%p, invalidating the connection...\n"),
                    pwrb_handle->pio_handle, pwrb_handle->pio_handle->ptarget));

                ap_complete_logout(pcontroller, 
                                pwrb_handle->plogout_handle,
                                ISCSI_PDU_LOGOUT_RESPONSE_FAIL, // not used if timeout
                                ACIT_TRUE);  // timeout = Yes
                break;

            case HWH_TYPE_TMF:

                ACIT_SET_TARGET_SUBSTATE(pwrb_handle->ptmf_handle->ptarget, 
                                        TARGET_SUB_STATE_TMF_ERROR);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on TMFH 0x%p in target 0x%p, will trigger session recovery...\n"),
                    pwrb_handle->ptmf_handle, pwrb_handle->ptmf_handle->ptarget));

                iscsi_recover_session(pcontroller, 
                                    pwrb_handle->ptmf_handle->ptarget,
                                    ACIT_NULL);

                break;

            case HWH_TYPE_NOP:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("CEV error on NOPH 0x%p in target 0x%p timeout, invalidating...\n"),
                    pwrb_handle->pnop_handle, pwrb_handle->pnop_handle->ptarget));

                //
                // We don't send out NOPOuts on our own, so we don't expect valid
                // ITT values for NOP handles..
                //
                OSM_ASSERT(0);
                phwi_controller->pacit_ws->partition[message_id].interrupt_stats.unknown_cmd_errors++;

                break;

            default:

                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error - unknown WRB 0x%x on CID 0x%x...\n", wrb_index, cid));
                OSM_ASSERT(0);
                phwi_controller->pacit_ws->partition[message_id].interrupt_stats.unknown_cmd_errors++;
                break;
            }
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Ignoring CQ Error on already invalidated cmd\n"));
        }
    }
    else
    {
        PACIT_TARGET_WS ptarget;

        pcontroller->eh_stats.cq_errors_invalid++;

        ptarget = iscsi_get_target_from_cid(pcontroller, cid);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ CMD Error on invalid WRB 0x%x cid 0x%x! Forcing session recovery on target 0x%p\n",
                    wrb_index, cid, ptarget));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_INVALID_CQ_ERR);

        iscsi_recover_session(pcontroller, 
                              ptarget,
                              ACIT_NULL);
    }
}

/* HWI API end */

PWRB_HANDLE
hwi_alloc_wrb_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri)
{
    PACIT_DC_LIST           plink;
    PWRB_HANDLE             pwrb_handle = ACIT_NULL;
    PHWI_WRB_CONTEXT        pwrb_context;

    pwrb_context = &phwi_controller->pwrb_context[cri];

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&pwrb_context->wrb_handle_list));

    LIST_REMOVEFROM_HEAD (&plink, 
                            &pwrb_context->wrb_handle_list);

    pwrb_handle = STRUCT_BASE(WRB_HANDLE,
                                link, 
                                plink);

    OSM_ASSERT(pwrb_handle);
    //pwrb_handle->count = 0;
    pwrb_context->wrb_handles_available--;


    if(!pwrb_context->wrb_handles_available)
    {
        /*
         * No more WRB handles available, this might be a case of
         * target going busy, let iscsi module handle this..
         */
        iscsi_flag_connection(phwi_controller->pacit_ws, 
                              cri, 
                              CONNECTION_BUSY_WRB_HANDLE,
                              ACIT_FALSE);  // ready = FALSE
    }

    return pwrb_handle;
}

ACIT_VOID
hwi_free_wrb_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri,
    PWRB_HANDLE             pwrb_handle)
{
    PHWI_WRB_CONTEXT        pwrb_context = &phwi_controller->pwrb_context[cri];

    /* Add to free list */
#ifdef RECYCLE_WRB_INDICES
    LIST_ADDTO_HEAD(&pwrb_handle->link, &pwrb_context->wrb_handle_list);
#else
    LIST_ADDTO_TAIL(&pwrb_handle->link, &pwrb_context->wrb_handle_list);
#endif

    if(pwrb_context->wrb_handles_available == 0)
    {
        iscsi_flag_connection(phwi_controller->pacit_ws,
                              cri,
                              CONNECTION_BUSY_WRB_HANDLE,
                              ACIT_TRUE);   // ready = TRUE
    }

    pwrb_context->wrb_handles_available++;

}

PSGL_HANDLE
hwi_alloc_io_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller)
{
    PACIT_DC_LIST           plink;
    PSGL_HANDLE             psgl_handle = ACIT_NULL;

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&phwi_controller->io_sgl_list));

    LIST_REMOVEFROM_HEAD (&plink, 
                          &phwi_controller->io_sgl_list);

    psgl_handle = STRUCT_BASE(SGL_HANDLE,
                                link, 
                                plink);

    OSM_ASSERT(psgl_handle);

    phwi_controller->io_sgl_handles_available--;

   if(!phwi_controller->io_sgl_handles_available)
    {
        /*
         * No more IO SGL handles available, notify OSM that controller is busy
         */
        ap_flag_busy_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_IO_SGL_HANDLE);
    }

    return psgl_handle;
}

ACIT_VOID
hwi_free_io_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    PSGL_HANDLE             psgl_handle)
{
#ifdef RECYCLE_SGL_INDICES
    LIST_ADDTO_HEAD(&psgl_handle->link, &phwi_controller->io_sgl_list);
#else
    LIST_ADDTO_TAIL(&psgl_handle->link, &phwi_controller->io_sgl_list);
#endif

   if(!phwi_controller->io_sgl_handles_available)
    {
        /*
         * Mark controller as ready
         */
        ap_flag_ready_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_IO_SGL_HANDLE);
    }

    phwi_controller->io_sgl_handles_available++;
}

PSGL_HANDLE
hwi_alloc_eh_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller)
{
    PACIT_DC_LIST           plink;
    PSGL_HANDLE             psgl_handle = ACIT_NULL;

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&phwi_controller->eh_sgl_list));

    LIST_REMOVEFROM_HEAD (&plink, 
                          &phwi_controller->eh_sgl_list);

    psgl_handle = STRUCT_BASE(SGL_HANDLE,
                                link, 
                                plink);

    OSM_ASSERT(psgl_handle);

    phwi_controller->eh_sgl_handles_available--;

   if(!phwi_controller->eh_sgl_handles_available)
    {
        /*
         * No more IO SGL handles available, notify OSM that controller is busy
         */
        ap_flag_busy_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_EH_SGL_HANDLE);
    }

    return psgl_handle;
}

ACIT_VOID
hwi_free_eh_sgl_handle(
    PHWI_CONTROLLER_WS      phwi_controller,
    PSGL_HANDLE             psgl_handle)
{
#ifdef RECYCLE_SGL_INDICES
    LIST_ADDTO_HEAD(&psgl_handle->link, &phwi_controller->eh_sgl_list);
#else
    LIST_ADDTO_TAIL(&psgl_handle->link, &phwi_controller->eh_sgl_list);
#endif

   if(!phwi_controller->eh_sgl_handles_available)
    {
        /*
         * Mark controller as ready
         */
        ap_flag_ready_controller(phwi_controller->pacit_ws, 
                                CONTROLLER_BUSY_EH_SGL_HANDLE);
    }

    phwi_controller->eh_sgl_handles_available++;
}

ACIT_VOID
hwi_get_wrb_mem(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cid,
    PACIT_UEXACT32          pwrb_mem_index,
    PACIT_UEXACT32          poffset,
    PACIT_UEXACT32          psize)
{
    *pwrb_mem_index = cid / phwi_controller->max_cxns_per_descriptor;
    *poffset = (cid % phwi_controller->max_cxns_per_descriptor) * phwi_controller->wrb_bytes_per_cxn;
    *psize = phwi_controller->wrb_bytes_per_cxn;
    *pwrb_mem_index += phwi_controller->wrb_mem_index;
}

ACIT_UEXACT32
hwi_write_be2_wrb_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           sgl_index)
{
    OSM_ASSERT(pwrb);
    OSM_ASSERT(pbhs_pa);

    //
    // This routine only updates the WRB. No SGL is assumed present.
    // NOPOuts in response to NOPIns will use this routine
    //

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in sgl_icd_index
    pwrb->sgl_icd_idx = sgl_index;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_write_be4_wrb_sgl(
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           sgl_index)
{
    OSM_ASSERT(pwrb);
    OSM_ASSERT(pbhs_pa);

    //
    // This routine only updates the WRB. No SGL is assumed present.
    // NOPOuts in response to NOPIns will use this routine
    //

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in sgl_index
    pwrb->sgl_idx = sgl_index;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_write_be2_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    OSM_ASSERT(pwrb);
    OSM_ASSERT(psgl_handle);
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in the data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);

            pwrb->sge0_last = 1;
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);
            ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);
        }

        if(pdata_sg_list->number_elements == 2)
            pwrb->sge1_last = 1;
    }

    // Fill in sgl_icd_index
    ACIT_PUT_LE32(pwrb->sgl_icd_idx, psgl_handle->sgl_index);

    _hwi_write_sgl(pbhs_pa, bhs_len, psgl_handle, pdata_sg_list);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_write_be4_sgl(
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    OSM_ASSERT(pwrb);
    OSM_ASSERT(psgl_handle);
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in the data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);

            pwrb->sge0_last = 1;
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);
            ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);
        }

        if(pdata_sg_list->number_elements == 2)
            pwrb->sge1_last = 1;
    }

    // Fill in sgl_index
    ACIT_PUT_LE32(pwrb->sgl_idx, psgl_handle->sgl_index);

    _hwi_write_sgl(pbhs_pa, bhs_len, psgl_handle, pdata_sg_list);

    return ACIT_STATUS_SUCCESS;
}

#ifdef ACIT_CFG_CHAINED_SGL

ACIT_VOID
_hwi_fill_chained_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_DC_LIST           psgh_list,
    PSGL_HANDLE             psgl_handle,
    PACIT_DC_LIST           list_position,
    PACIT_SG_LIST           pdata_sg_list)
{
    ISCSI_SGE               *sge;
    PACIT_SG_ELEMENT        acit_sge;
    ACIT_UEXACT32           i, j, offset = 0;
    ACIT_UEXACT32           num_sges_in_sgl;
    ACIT_UEXACT32           primary = 1, chained = 0, force_chain = 0;

    //
    // The first FR must still point to the area where the
    // PDU templates are located
    //
    sge = (PISCSI_SGE)psgl_handle->pfrag;
    FILL_SGE(sge, bhs_len, pbhs_pa->u.a32.address_lo, pbhs_pa->u.a32.address_hi, 0, 0);
       
    // Check if the command has additional SGE's
    if (pdata_sg_list && pdata_sg_list->number_elements) 
    {
        sge++;

        // In chained SGL, the skip SGE has to be explicitly programmed
        FILL_SGE(sge, 0, 0, 0, 0, 0);

        //
        // Get pointer to the ACIT scatter gather list and fill
        // in rest of the FRs
        //

        //
        // do a increment of SGE before we start in order to skip over the first
        // fragment
        //
        ++sge;

        // In the first SGL, 2 are reserved. One is BHS and the other is skip SGE.
        // In addition, the last one is reserved for chain. So initialize it to -3
        // Later, the loops will initialize it to num_sge - 1. -1 because one SGE
        // is taken up for the chain
        num_sges_in_sgl = pcontroller->params.num_sge - 3;

        if(pdata_sg_list->number_elements < num_sges_in_sgl)
        {
            num_sges_in_sgl = pdata_sg_list->number_elements;
        }

        acit_sge = pdata_sg_list->sglist;
        i=0;

        do
        {
            for(j = 0; j < num_sges_in_sgl; j++)
            {
                OSM_ASSERT (acit_sge->length > 0);
                OSM_ASSERT (acit_sge->address.u.a64.address != 0);
                OSM_ASSERT (offset < pio_handle->exp_xfer_length);

                // !!Warn!!
                // For DMA tests, each SGEs are already setup for <64K
                // So no checks are needed

                FILL_SGE(sge,
                         acit_sge->length,
                         acit_sge->address.u.a32.address_lo,
                         acit_sge->address.u.a32.address_hi,
                         offset,
                         0);

                i++;
                offset+=sge->len;
                ++acit_sge;
                ++sge;

                // For primary ICD, after filling upto chain_at elements,
                // force a break
                if(primary && 
                   (j >= 2) && 
                   (j > pcontroller->dma_test.chain_at) &&
                   (j < DMA_TEST_FRAG_NUM_MAX_START_OFFSET) &&
                   (j != (num_sges_in_sgl - 1)))
                {
                    force_chain = 1;
                    break;
                }
            }

            if(force_chain || i < pdata_sg_list->number_elements)
            {
                PSGL_HANDLE temp;
                ISCSI_SGL_CRA_V2 cra;
                PISCSI_CHAIN_SGE_V2 chain_sge = (PISCSI_CHAIN_SGE_V2)sge;
                cra.dw = 0;

                // If we've not filled all acit_sges, then we need to chain the SGLs.
                // Fill a chain SGE and pick the next list. In order to fill in the
                // chain, we need the next member now. So get that first.

                list_position = list_position->next;
                if(list_position == psgh_list)
                {
                    // end of list. Break. But if we've not filled each of the acit_sges,
                    // then that means we had not allocated enough to begin with.
                    // Assert on that condition for now. We have to fail the IO
#if ACIT_CFG_DEBUG_BUILD
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("!!!Critical Driver Bug!!! Can cause possible data corruption*************\n"));
                    OSM_ASSERT(0);
#else
                    ap_crash_system(ACIT_NULL);
#endif
                    break;
                }

                temp = psgl_handle;
                psgl_handle = STRUCT_BASE(SGL_HANDLE, link, list_position);

                // Initialize the parameters for the next loop
                // One at the end is reserved for the chain element
                num_sges_in_sgl = (pdata_sg_list->number_elements - i);

                // Note!
                // The frag_num offset is only for the first chain element
                // After that, all chains will have the frag_num offset set to 0
                //
                if((chained == 0) &&
                   ((pcontroller->dma_test.test_case == DMA_TEST_INC_FRAG_NUM) ||
                    (pcontroller->dma_test.test_case == DMA_TEST_SEQINC_FRAG_NUM)))
                {
                    // Remaining is the remaining number of SGEs
                    // Effective is the number of SGEs that can be filled based on the frag_num offset
                    ACIT_UEXACT32 remain, effective;

                    remain = num_sges_in_sgl; // already calculated above

                    if(pcontroller->dma_test.chain_at >= DMA_TEST_FRAG_NUM_MAX_START_OFFSET)
                        pcontroller->dma_test.chain_at = 0;

                    if(pcontroller->dma_test.test_case == DMA_TEST_INC_FRAG_NUM)
                    {
                        if(((pcontroller->stats.ios_submitted + 1) % DMA_TEST_INC_FRAG_NUM_AT) == 0)
                        {
                            pcontroller->dma_test.chain_at++;
                            if(pcontroller->dma_test.chain_at == 1)
                            {
                                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Reset frag_num offset to 1\n"));
                            }
                        }
                    }
                    else if(pcontroller->dma_test.test_case == DMA_TEST_SEQINC_FRAG_NUM)
                    {
                        pcontroller->dma_test.chain_at++;
                    }

                    cra.frag_num = pcontroller->dma_test.chain_at;

                    effective = (pcontroller->params.num_sge - 1) - cra.frag_num;
                    num_sges_in_sgl = ACIT_MIN(remain, effective);
                    OSM_ASSERT(num_sges_in_sgl);

                    // Init start of SGE in next SGL based on the frag_num offset
                    sge = (PISCSI_SGE)psgl_handle->pfrag + cra.frag_num;
                }
                else
                {
                    // Init start of SGE in next SGL
                    sge = (PISCSI_SGE)psgl_handle->pfrag;
                }

                if(num_sges_in_sgl > (pcontroller->params.num_sge - 1))
                {
                    num_sges_in_sgl = (pcontroller->params.num_sge - 1);
                }

                OSM_ASSERT((num_sges_in_sgl + i) <= pdata_sg_list->number_elements);

                cra.sgl_index = psgl_handle->sgl_index;
                cra.constant = 0x2;
                FILL_CHAIN_SGE(chain_sge, cra);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("IOH 0x%p Chaining SGL 0x%x (0x%p) to SGL 0x%x (0x%p), cra 0x%x\n",
                    pio_handle, psgl_handle->sgl_index, psgl_handle, temp->sgl_index, temp, cra.dw));

                chained = 1; // Set this to indicate a chain is already formed
                force_chain = 0; // Don't force anymore chains
                primary = 0; // whatever follows is secondary ICD
            }

        } while(i < pdata_sg_list->number_elements);

        sge--;
        sge->last_sge = 1;

        {
            PISCSI_CHAIN_SGE_V2 check_sge = (PISCSI_CHAIN_SGE_V2)sge;
            if((check_sge->type != 0) && (check_sge->last == 1))
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Chained element marked last SGE! Driver Bug!!!! 0x%p 0x%p\n",
                    pio_handle, check_sge));
                OSM_ASSERT(0);
            }
        }
    }
}


ACIT_UEXACT32
get_frag_num_from_position(ACIT_UEXACT32 pos)
{
    // pos is 1-based, return value is 0-based

    ACIT_UEXACT32 i = pos/4;

    switch(i) {
        case 0: OSM_ASSERT(0); return 0xff;
        case 1: OSM_ASSERT(0); return 0xff;
        case 2: return 7*4;
        case 3: return 6*4;
        case 4: return 5*4;
        case 5: return 0*4;
        case 6: return 4*4;
        case 7: return 3*4;
        case 8: return 2*4;
        default: OSM_ASSERT(0); return 0xff;
    }

}
//
// This function does not have to handle the 64K split since it's already
// split at 64K by ap_init_dmatest_sg_list
//
ACIT_VOID
_hwi_fill_chained_sgl_multi_chain(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_DC_LIST           psgh_list,
    PSGL_HANDLE             psgl_handle,
    PACIT_DC_LIST           list_position,
    PACIT_SG_LIST           pdata_sg_list)
{
    ISCSI_SGE               *sge;
    PACIT_SG_ELEMENT        acit_sge;
    ACIT_UEXACT32           i, j, offset = 0;
    ACIT_UEXACT32           num_sges_in_sgl;
    ACIT_UEXACT32           frag_num; // is 1-based
    ACIT_UEXACT32           num_frag_total = 0;

    //
    // The first FR must still point to the area where the
    // PDU templates are located
    //
    sge = (PISCSI_SGE)psgl_handle->pfrag;
    FILL_SGE(sge, bhs_len, pbhs_pa->u.a32.address_lo, pbhs_pa->u.a32.address_hi, 0, 0);
       
    // Check if the command has additional SGE's
    if (pdata_sg_list && pdata_sg_list->number_elements) 
    {
        sge++;

        // In chained SGL, the skip SGE has to be explicitly programmed
        FILL_SGE(sge, 0, 0, 0, 0, 0);

        //
        // Get pointer to the ACIT scatter gather list and fill
        // in rest of the FRs
        //

        //
        // do a increment of SGE before we start in order to skip over the first
        // fragment
        //
        ++sge;

        frag_num = 3;

        // In the first SGL, 2 are reserved. One is BHS and the other is skip SGE.
        // In addition, the last one is reserved for chain. So initialize it to -3
        // Later, the loops will initialize it to num_sge - 1. -1 because one SGE
        // is taken up for the chain
        num_sges_in_sgl = pcontroller->params.num_sge - 3;

        if(pdata_sg_list->number_elements < num_sges_in_sgl)
        {
            num_sges_in_sgl = pdata_sg_list->number_elements;
        }

        acit_sge = pdata_sg_list->sglist;
        i=0;

        do
        {
            for(j = 0; j < num_sges_in_sgl; j++)
            {
                OSM_ASSERT (acit_sge->length > 0);
                OSM_ASSERT (acit_sge->address.u.a64.address != 0);
                OSM_ASSERT (offset < pio_handle->exp_xfer_length);

                // !!!Warn!!!
                // Do not start chaining when frag_num is 4. Because then the WRB itself
                // should contain the chain element
                if((frag_num % 4 == 0) &&
                    (frag_num != 4))
                {
                    ISCSI_SGL_CRA_V2 cra;
                    PISCSI_CHAIN_SGE_V2 chain_sge = (PISCSI_CHAIN_SGE_V2)sge;

                    cra.dw = 0;

                    // Chain here
                    cra.sgl_index = psgl_handle->sgl_index;
                    cra.constant = 0x2;
                    cra.frag_num = get_frag_num_from_position(frag_num);

                    if(cra.frag_num == 0)
                    {
                        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("Returning 0 for position %d total frags %d\n", frag_num, num_frag_total));
                        break;
                    }

                    OSM_ASSERT(cra.frag_num != 0xFF);
                    FILL_CHAIN_SGE(chain_sge, cra);
                    num_frag_total++;

                    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("IOH 0x%p SGLH 0x%p Chaining SGL at frag_num 0x%x to frag_num 0x%x, cra 0x%x\n",
                        pio_handle, psgl_handle, frag_num-1, cra.frag_num, cra.dw));
 
                    // Setup next pointer for SGE and next frag_num
                    sge = psgl_handle->pfrag + cra.frag_num;
                    frag_num = cra.frag_num + 1;
                }

                //
                // Warn!!
                // For DMA tests, each SGEs are already setup for <64K
                // So no checks are needed
                FILL_SGE(sge,
                         acit_sge->length,
                         acit_sge->address.u.a32.address_lo,
                         acit_sge->address.u.a32.address_hi,
                         offset,
                         0);

                i++;
                ++sge;
                offset+=acit_sge->length;
                ++acit_sge;
                frag_num++;
                num_frag_total++;
            }

            if(i < pdata_sg_list->number_elements)
            {
                PSGL_HANDLE temp;
                // If we've not filled all acit_sges, then we need to chain the SGLs.
                // Fill a chain SGE and pick the next list. In order to fill in the
                // chain, we need the next member now. So get that first.

                ISCSI_SGL_CRA_V2 cra;
                PISCSI_CHAIN_SGE_V2 chain_sge = (PISCSI_CHAIN_SGE_V2)sge;

                cra.dw = 0;

                list_position = list_position->next;
                if(list_position == psgh_list)
                {
                    // end of list. Break. But if we've not filled each of the acit_sges,
                    // then that means we had not allocated enough to begin with.
                    // Assert on that condition for now. We have to fail the IO
#if ACIT_CFG_DEBUG_BUILD
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("!!!Critical Driver Bug!!! Can cause possible data corruption*************\n"));
                    OSM_ASSERT(0);
#else
                    ap_crash_system(ACIT_NULL);
#endif
                    break;
                }

                temp = psgl_handle;
                psgl_handle = STRUCT_BASE(SGL_HANDLE, link, list_position);

                // Init start of SGE in next SGL
                sge = (PISCSI_SGE)psgl_handle->pfrag;

                // Initialize the parameters for the next loop
                // One at the end is reserved for the chain element
                num_sges_in_sgl = (pdata_sg_list->number_elements - i);
                if(num_sges_in_sgl > (pcontroller->params.num_sge - 1))
                {
                    num_sges_in_sgl = (pcontroller->params.num_sge - 1);
                }

                cra.sgl_index = psgl_handle->sgl_index;
                cra.constant = 0x2;
                FILL_CHAIN_SGE(chain_sge, cra);

                num_frag_total = 0;
                frag_num = 1;

                OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("IOH 0x%p Chaining SGL 0x%x (0x%p) to SGL 0x%x (0x%p), cra 0x%x\n",
                    pio_handle, psgl_handle->sgl_index, psgl_handle, temp->sgl_index, temp, cra.dw));
            }

        } while(i < pdata_sg_list->number_elements);

        sge--;
        sge->last_sge = 1;

        {
            PISCSI_CHAIN_SGE_V2 check_sge = (PISCSI_CHAIN_SGE_V2)sge;
            if((check_sge->type != 0) && (check_sge->last == 1))
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Chained element marked last SGE! Driver Bug!!!! 0x%p 0x%p\n",
                    pio_handle, check_sge));
                OSM_ASSERT(0);
            }
        }
    }
}

ACIT_UEXACT32
hwi_write_be2_chained_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_DC_LIST           psgh_list,
    PACIT_SG_LIST           pdata_sg_list)
{
    ISCSI_SGE               *sge;
    PACIT_SG_ELEMENT        acit_sge;
    ACIT_UEXACT32           i, offset = 0;
    PSGL_HANDLE             psgl_handle;
    PACIT_DC_LIST           list_position;

    OSM_ASSERT(pwrb);
    //OSM_ASSERT(psgl_handle);
    OSM_ASSERT(IS_LIST_NOT_EMPTY(psgh_list));
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    list_position = psgh_list->next;
    psgl_handle = STRUCT_BASE(SGL_HANDLE, link, list_position);
    OSM_ASSERT(psgl_handle);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);

            if(pdata_sg_list->sglist[0].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge0_len, BE_MAX_BYTES_PER_FRAG);

                ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo +
                                                    BE_MAX_BYTES_PER_FRAG);
                ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
                ACIT_PUT_LE32(pwrb->sge1_r2t_offset, BE_MAX_BYTES_PER_FRAG);

                if((pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG) > BE_MAX_BYTES_PER_FRAG)
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
                }
                else
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG);
                }

                goto fill_rest;
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);
                pwrb->sge0_last = 1;
            }
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);

            if(pdata_sg_list->sglist[1].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);

                if(pdata_sg_list->number_elements == 2)
                    pwrb->sge1_last = 1;
            }
        }
    }

fill_rest:

    // Fill in sgl_icd_index
    ACIT_PUT_LE32(pwrb->sgl_icd_idx, psgl_handle->sgl_index);

    _hwi_fill_chained_sgl(pcontroller,
                          pio_handle,
                          pbhs_pa,
                          bhs_len,
                          psgh_list,
                          psgl_handle,
                          list_position,
                          pdata_sg_list);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_write_be4_chained_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PACIT_DC_LIST           psgh_list,
    PACIT_SG_LIST           pdata_sg_list)
{
    ISCSI_SGE               *sge;
    PACIT_SG_ELEMENT        acit_sge;
    ACIT_UEXACT32           i, offset = 0;
    PSGL_HANDLE             psgl_handle;
    PACIT_DC_LIST           list_position;

    OSM_ASSERT(pwrb);
    //OSM_ASSERT(psgl_handle);
    OSM_ASSERT(IS_LIST_NOT_EMPTY(psgh_list));
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    list_position = psgh_list->next;
    psgl_handle = STRUCT_BASE(SGL_HANDLE, link, list_position);
    OSM_ASSERT(psgl_handle);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);

            if(pdata_sg_list->sglist[0].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge0_len, BE_MAX_BYTES_PER_FRAG);

                ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo +
                                                    BE_MAX_BYTES_PER_FRAG);
                ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
                ACIT_PUT_LE32(pwrb->sge1_r2t_offset, BE_MAX_BYTES_PER_FRAG);

                if((pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG) > BE_MAX_BYTES_PER_FRAG)
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
                }
                else
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG);
                }

                goto fill_rest;
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);
                pwrb->sge0_last = 1;
            }
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);

            if(pdata_sg_list->sglist[1].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);

                if(pdata_sg_list->number_elements == 2)
                    pwrb->sge1_last = 1;
            }
        }
    }

fill_rest:

    // Fill in sgl_icd_index
    ACIT_PUT_LE32(pwrb->sgl_idx, psgl_handle->sgl_index);

    if(pcontroller->dma_test.test_case == DMA_TEST_MULTIPLE_CHAINS_PER_SGL)
    {
        _hwi_fill_chained_sgl_multi_chain(pcontroller,
                              pio_handle,
                              pbhs_pa,
                              bhs_len,
                              psgh_list,
                              psgl_handle,
                              list_position,
                              pdata_sg_list);
    }
    else
    {
        _hwi_fill_chained_sgl(pcontroller,
                              pio_handle,
                              pbhs_pa,
                              bhs_len,
                              psgh_list,
                              psgl_handle,
                              list_position,
                              pdata_sg_list);
    }

    return ACIT_STATUS_SUCCESS;
}

#endif

ACIT_UEXACT32
hwi_write_be2_large_sgl(
    PISCSI_WRB              pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    OSM_ASSERT(pwrb);
    OSM_ASSERT(psgl_handle);
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);

            if(pdata_sg_list->sglist[0].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge0_len, BE_MAX_BYTES_PER_FRAG);

                ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo +
                                                    BE_MAX_BYTES_PER_FRAG);
                ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
                ACIT_PUT_LE32(pwrb->sge1_r2t_offset, BE_MAX_BYTES_PER_FRAG);

                if((pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG) > BE_MAX_BYTES_PER_FRAG)
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
                }
                else
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG);
                }

                goto fill_rest;
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);
                pwrb->sge0_last = 1;
            }
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);

            if(pdata_sg_list->sglist[1].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);

                if(pdata_sg_list->number_elements == 2)
                    pwrb->sge1_last = 1;
            }
        }
    }

fill_rest:

    // Fill in sgl_icd_index
    ACIT_PUT_LE32(pwrb->sgl_icd_idx, psgl_handle->sgl_index);

    _hwi_write_large_sgl(pbhs_pa, bhs_len, psgl_handle, pdata_sg_list);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_write_be4_large_sgl(
    PISCSI_SKH_WRB_V2       pwrb,
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    OSM_ASSERT(pwrb);
    OSM_ASSERT(psgl_handle);
    OSM_ASSERT(pbhs_pa);
    OSM_ASSERT(bhs_len);

    // Fill in BHS portion of WRB
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_lo, pbhs_pa->u.a32.address_lo);
    ACIT_PUT_LE32(pwrb->iscsi_bhs_addr_hi, pbhs_pa->u.a32.address_hi);

    // Fill in data SG1 and SG2
    if(pdata_sg_list)
    {
        if(pdata_sg_list->number_elements >= 1)
        {
            ACIT_PUT_LE32(pwrb->sge0_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge0_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);

            if(pdata_sg_list->sglist[0].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge0_len, BE_MAX_BYTES_PER_FRAG);

                ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[0].address.u.a32.address_lo +
                                                    BE_MAX_BYTES_PER_FRAG);
                ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[0].address.u.a32.address_hi);
                ACIT_PUT_LE32(pwrb->sge1_r2t_offset, BE_MAX_BYTES_PER_FRAG);

                if((pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG) > BE_MAX_BYTES_PER_FRAG)
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
                }
                else
                {
                    ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[0].length - BE_MAX_BYTES_PER_FRAG);
                }

                goto fill_rest;
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge0_len, pdata_sg_list->sglist[0].length);
                pwrb->sge0_last = 1;
            }
        }

        if(pdata_sg_list->number_elements >= 2)
        {
            pwrb->sge0_last = 0;
            ACIT_PUT_LE32(pwrb->sge1_addr_lo, pdata_sg_list->sglist[1].address.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->sge1_addr_hi, pdata_sg_list->sglist[1].address.u.a32.address_hi);
            ACIT_PUT_LE32(pwrb->sge1_r2t_offset, pdata_sg_list->sglist[0].length);

            if(pdata_sg_list->sglist[1].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_PUT_LE32(pwrb->sge1_len, BE_MAX_BYTES_PER_FRAG);
            }
            else
            {
                ACIT_PUT_LE32(pwrb->sge1_len, pdata_sg_list->sglist[1].length);

                if(pdata_sg_list->number_elements == 2)
                    pwrb->sge1_last = 1;
            }
        }
    }

fill_rest:

    // Fill in sgl_icd_index
    ACIT_PUT_LE32(pwrb->sgl_idx, psgl_handle->sgl_index);

    _hwi_write_large_sgl(pbhs_pa, bhs_len, psgl_handle, pdata_sg_list);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_init_wrb_handle(
    PACIT_CONTROLLER_WS      pcontroller)
{
    PHWI_CONTROLLER_WS      phwi;
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    ACIT_UEXACT32           wrbh_desc, wrb_desc;
    ACIT_UEXACT32           wrb_index;
    ACIT_UEXACT32           index, wrbh_total, wrb_total, cxns;
    PWRB_HANDLE             pwrb_handle = ACIT_NULL;
    PISCSI_WRB              pwrb = ACIT_NULL;
    PHWI_WRB_CONTEXT        pwrb_context;

    // If ever the size of WRB changes for different hardware, then this needs to
    // revisited

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    phwi->pwrb_context = (PHWI_WRB_CONTEXT)GET_MEM_DESC_VA(pc, HWI_MEM_WRB_CONTEXT);

    /*
     * Split the WRB handles across the pwrb_handle_base
     */
    wrbh_desc = phwi->wrbh_mem_index;
    OSM_ASSERT(wrbh_desc == pc->mem_descriptors[wrbh_desc].index);
    pwrb_handle = (PWRB_HANDLE)pc->mem_descriptors[wrbh_desc].virtual_address;

    wrb_desc = phwi->wrb_mem_index;
    OSM_ASSERT(wrb_desc == pc->mem_descriptors[wrb_desc].index);
    pwrb = (PISCSI_WRB)pc->mem_descriptors[wrb_desc].virtual_address;

    wrbh_total = 0;
    wrb_total = 0;
    cxns = 0;

    for(index = 0; index < p->num_cxn; index++)
    {
        pwrb_context = &phwi->pwrb_context[index];

        LIST_INIT_HEADER(&pwrb_context->wrb_handle_list);
        LIST_INIT_HEADER(&pwrb_context->wrb_handle_drvr_list);
        pwrb_context->plast_wrb = ACIT_NULL;

        pwrb_context->pwrb_handle_base = pwrb_handle;

        // Initialize this to 0 since after a hibernation resume this is not 
        // guaranteed to be 0
        pwrb_context->wrb_handles_available = 0;

        for(wrb_index = 0; wrb_index < p->num_cmd_per_cxn; wrb_index++)
        {
            pwrb_handle->wrb_index = (ACIT_UEXACT16)wrb_index;
            pwrb_handle->pbe2_wrb = pwrb;

            LIST_ADDTO_TAIL(&pwrb_handle->link,
                            &pwrb_context->wrb_handle_list);

            pwrb_context->wrb_handles_available++;

            pwrb_handle ++;
            pwrb++;
            OSM_ASSERT(wrbh_total * sizeof(WRB_HANDLE) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
            OSM_ASSERT(wrb_total * sizeof(ISCSI_WRB) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
            wrbh_total ++;
            wrb_total ++;
        }

        cxns++;

        if(index < p->num_cxn - 1)
        {
            if(wrbh_total  >= phwi->max_wrbh_per_descriptor)
            {
                wrbh_desc++;
                pwrb_handle = (PWRB_HANDLE)pc->mem_descriptors[wrbh_desc].virtual_address;
                wrbh_total = 0;
            }

            if(cxns >= phwi->max_cxns_per_descriptor)
            {
                wrb_desc++;
                pwrb = (PISCSI_WRB)pc->mem_descriptors[wrb_desc].virtual_address;
                wrb_total = 0;
                cxns = 0;
            }
        }
    }
}

ACIT_VOID
hwi_init_sgl_handle(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi;
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    ACIT_UEXACT16           index, sgl_total, sgl_desc, sglh_total;
    PISCSI_SGE              pfrag;
    PSGL_HANDLE             psgl_handle;
    ACIT_UEXACT32           ulp, ulp_icd_index, num_icd;
    ACIT_UEXACT32           icd_begin, icd_end; // True range reported by FW
    ACIT_UEXACT32           aligned_begin, aligned_count; // Ranges aligned to page boundary

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    psgl_handle = (PSGL_HANDLE)pc->mem_descriptors[HWI_MEM_SGLH].virtual_address;

    sgl_desc = (ACIT_UEXACT16)phwi->frag_mem_index;
    OSM_ASSERT(sgl_desc == pc->mem_descriptors[sgl_desc].index);
    pfrag = (PISCSI_SGE)pc->mem_descriptors[sgl_desc].virtual_address;

    LIST_INIT_HEADER(&phwi->io_sgl_list);
    LIST_INIT_HEADER(&phwi->eh_sgl_list);

    phwi->psgl_handle_base = psgl_handle;
    phwi->io_sgl_handles_available = 0;
    phwi->eh_sgl_handles_available = 0;

    sgl_total = 0;
    sglh_total = 0;

    // Bug 138859 : The combined icd_count and icd range is reported on each Ulp config
    ulp = ap_get_initial_ulp(pcontroller);

    // Bug 147561 : Use page-aligned icd ranges.
    icd_begin = pc->fw_config.ulp[ulp].icd_start;
    num_icd = pc->fw_config.ulp[ulp].icd_count;
    icd_end = icd_begin + num_icd - 1; 

    aligned_begin = pc->fw_config.aligned_icd_start;
    aligned_count = pc->fw_config.aligned_icd_count;

    ulp_icd_index = 0;

    for(index = 0; index < p->icds_per_controller; index++)
    {
        psgl_handle->sgl_index = aligned_begin + ulp_icd_index;

        if((aligned_begin + ulp_icd_index) > (aligned_begin + p->icds_per_controller - 1))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                ("Incorrect SGL Index being initialized : 0x%x Range - [0x%x:0x%x]\n",
                (aligned_begin + ulp_icd_index),
                aligned_begin,
                aligned_begin + p->icds_per_controller - 1));
            OSM_ASSERT(0);  // Do NOT ignore
        }

        psgl_handle->pfrag = pfrag;

        if(index < p->ios_per_controller)
        {
            LIST_ADDTO_TAIL(&psgl_handle->link,
                            &phwi->io_sgl_list);

            phwi->io_sgl_handles_available++;
        }
        else
        {
            LIST_ADDTO_TAIL(&psgl_handle->link,
                            &phwi->eh_sgl_list);

            phwi->eh_sgl_handles_available++;
        }

        psgl_handle ++;
        pfrag = (PISCSI_SGE)((ACIT_PTR)pfrag + (p->num_sge * sizeof(ISCSI_SGE)));
        sgl_total ++;
        sglh_total ++;
        ulp_icd_index++;

        if(sgl_total >= phwi->max_sgl_per_descriptor)
        {
            if(index < p->icds_per_controller - 1)
            {
                sgl_desc++;
                pfrag = (PISCSI_SGE)pc->mem_descriptors[sgl_desc].virtual_address;
                sgl_total = 0;
            }
            else
            {
                break;
            }
        }
        OSM_ASSERT(sgl_total * sizeof(ISCSI_SGE) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
        OSM_ASSERT(sglh_total * sizeof(SGL_HANDLE) <= ACIT_CFG_MAX_CONTIGUOUS_MEM);
    }

    hwi_unitt_init_sgl_handle(pcontroller);
}

ACIT_UEXACT32
hwi_reserve_nop_sgl(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi;
    PSGL_HANDLE             psgl_handle;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    if(phwi->eh_sgl_handles_available)
    {
        psgl_handle = hwi_alloc_eh_sgl_handle(phwi);

        if(psgl_handle)
        {
            phwi->nop_icd = psgl_handle->sgl_index;
            return ACIT_STATUS_SUCCESS;
        }

        return ACIT_STATUS_ERROR;
    }
    else
    {
        // Do not return failure in the interest of allowing
        // driver load to succeed.
        return ACIT_STATUS_SUCCESS;
    }
}

ACIT_VOID
hwi_unitt_init_sgl_handle(
    PACIT_CONTROLLER_WS pcontroller)
{
    PHWI_CONTROLLER_WS  phwi;
    PSGL_HANDLE         psgl_handle;
    PACIT_DC_LIST       flink;
    ACIT_UEXACT32       num_handles = 0;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    if(phwi->io_sgl_handles_available)
    {
        OSM_ASSERT(IS_LIST_NOT_EMPTY(&phwi->io_sgl_list));

        flink = phwi->io_sgl_list.next;

        while(flink != &phwi->io_sgl_list)
        {
            psgl_handle = STRUCT_BASE(SGL_HANDLE,
                                      link,
                                      flink);

#if 0
            // Disabled for Dual Chute
            OSM_ASSERT(psgl_handle->sgl_index == num_handles +
                pcontroller->fw_config.icd_start);
#endif

            num_handles++;
            flink = flink->next;
        }
    }

    OSM_ASSERT(num_handles == phwi->io_sgl_handles_available);
}

INLINE
ACIT_VOID
_hwi_write_sgl(
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    PISCSI_SGE              sge;
    PACIT_SG_ELEMENT        acit_sge;
    ACIT_UEXACT32           i, offset = 0;

    //
    // The first FR must still point to the area where the
    // PDU templates are located
    //
    sge = (PISCSI_SGE)psgl_handle->pfrag;

    FILL_SGE(sge,
             bhs_len,
             pbhs_pa->u.a32.address_lo,
             pbhs_pa->u.a32.address_hi,
             0,
             0);
       
    // Check if the command has additional SGE's
    if (pdata_sg_list && pdata_sg_list->number_elements) 
    {
        sge++;

        //
        // Get pointer to the ACIT scatter gather list and fill
        // in rest of the FRs
        //

        acit_sge = pdata_sg_list->sglist;

        for(i = 0; i < pdata_sg_list->number_elements; i++)
        { 
            //
            // do a increment of SGL before we start in order to skip over the first
            // fragment
            //
            ++sge;

            OSM_ASSERT (acit_sge->length > 0);
            OSM_ASSERT (acit_sge->address.u.a64.address != 0);

            FILL_SGE(sge,
                     acit_sge->length,
                     acit_sge->address.u.a32.address_lo,
                     acit_sge->address.u.a32.address_hi,
                     offset,
                     0);

            offset+=sge->len;
            ++acit_sge;
        }

        sge->last_sge = 1;
    }
}

ACIT_VOID
_hwi_write_large_sgl(
    PACIT_BUS_ADDRESS       pbhs_pa,
    ACIT_UEXACT32           bhs_len,
    PSGL_HANDLE             psgl_handle,
    PACIT_SG_LIST           pdata_sg_list)
{
    PISCSI_SGE              sge;
    PACIT_SG_ELEMENT        acit_sge;
    ACIT_UEXACT32           i, offset = 0;

    //
    // The first FR must still point to the area where the
    // PDU templates are located
    //
    sge = (PISCSI_SGE)psgl_handle->pfrag;
    FILL_SGE(sge, bhs_len, pbhs_pa->u.a32.address_lo, pbhs_pa->u.a32.address_hi, 0, 0);
    // Check if the command has additional SGE's
    if (pdata_sg_list && pdata_sg_list->number_elements) 
    {
        sge++;

        //
        // Get pointer to the ACIT scatter gather list and fill
        // in rest of the FRs
        //

        acit_sge = pdata_sg_list->sglist;

        for(i = 0; i < pdata_sg_list->number_elements; i++)
        { 
            //
            // do a increment of SGE before we start in order to skip over the first
            // fragment
            //
            ++sge;

            OSM_ASSERT (acit_sge->length > 0);
            OSM_ASSERT (acit_sge->address.u.a64.address != 0);

            FILL_SGE(sge,
                     0,
                     acit_sge->address.u.a32.address_lo,
                     acit_sge->address.u.a32.address_hi,
                     offset,
                     0);
            //
            // The hardware has a limitation with DMA in that only 64KB or less
            // can be in a frag. Split anything that is more than this limit
            //
            if(acit_sge->length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_UEXACT32   bytes_remaining, len;
                ACIT_BUS_ADDRESS address;

                address.u.a32.address_lo = sge->addr_lo;
                address.u.a32.address_hi = sge->addr_hi;

                // limit the len to max allowed
                ACIT_PUT_LE32(sge->len, BE_MAX_BYTES_PER_FRAG);

                bytes_remaining = acit_sge->length - sge->len;

                do
                {
                    offset+=sge->len;
                    ++sge;

                    if(bytes_remaining > BE_MAX_BYTES_PER_FRAG)
                        len = BE_MAX_BYTES_PER_FRAG;
                    else
                        len = bytes_remaining;

                    FILL_SGE(sge,
                             len,
                             address.u.a32.address_lo + (acit_sge->length - bytes_remaining),
                             address.u.a32.address_hi,
                             offset,
                             0);

                    bytes_remaining = bytes_remaining - len;

                } while(bytes_remaining > 0);
            }
            else
            {
                ACIT_PUT_LE32(sge->len, acit_sge->length);
            }

            offset+=sge->len;
            ++acit_sge;
        }

        sge->last_sge = 1;
    }
}